/**
 * Constants
 */
var gc = gc || {};

var numPending = 0;
var maxNumPending = 0;

var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

const Cmd_SPI_Control = 0x08;

var getDeviceIndexSpi = function(coreIndex, regModelId) {
    coreIndex = coreIndex === undefined ? 0 : coreIndex;

    if(coreIndex > -1 && regModelId === 'bq79616') {
        coreIndex = coreIndex + 1;
    }

    return coreIndex;
};

var init = function () {

    var regReadDetail = undefined, regReadCount, regReadPromiseArr;

    /**
     * This code implements an example custom USB2ANY Interface (functionally just a clone of the default I2C interface)
     *
     * @constructor
     * @param settings - The interface settings defined in the system.json
     * @param registerModel - The register model
     * @implements gc.databind.USB2ANY.ICommInterface
     */
    CustomSPI_Interface = function (settings, registerModel) {
        this._registerModel = registerModel;
        this.transferPromises = [];
        this.cmdQueue = [];
        this.readyToProcess = true;
        this.noOfDevices = 0;
        this.config = {};
        this.is600 = false;
        this.opPromise;
    };

    /**
     * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
     * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
     * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
     */
    gc.databind.USB2ANY.prototype.CUSTOMSPI_Interface = CustomSPI_Interface;

    /**
     * Custom interfaces should be extended from the ICommInterface base class
     */
    CustomSPI_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

    /**
     * Set the communication parameters
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param settings - The interface settings defined in the system.json
     */
    CustomSPI_Interface.prototype.control = function (u2a, settings) {
        console.log('SPI', 'Control', u2a, settings);

        var clkPhase = settings.clkPhase === 'leadingEdge' ? 1 : 0;
        var clkPolarity = settings.clkPolarity === 'inactiveHigh' ? 1 : 0;
        var bitDir = settings.bitDir === 'MSBFirst' ? 1 : 0;
        var charLen = settings.charLen === 7 ? 1 : 0;
        var csType;
        switch(settings.csType) {
            case 'everyByte': csType = 0;
                break;
            case 'everyPacket': csType = 1;
                break;
            case 'everyWord': csType = 2;
                break;
            case 'noCS': csType = 3;
                break;
            case 'everyPacket_MOSIInactiveHigh': csType = 4;
                break;
            case 'pulseAfterPacket': csType = 255;
                break;
            default: csType = 0;
                break;
        }
        var csPolarity = settings.csPolarity === 'activeLow' ? 1 : 0;

        var divHigh = parseInt(settings.divHigh);
        if(isNaN(divHigh)) {
            divHigh = 0;
        } else {
            if(divHigh < 0) {
                divHigh = 0;
            } else if(divHigh > 255) {
                divHigh = 255;
            }
        }

        var divLow = parseInt(settings.divLow);
        if(isNaN(divLow)) {
            divLow = 0;
        } else {
            if(divLow < 0) {
                divLow = 0;
            } else if(divLow > 255) {
                divLow = 255;
            }
        }

        if(divHigh === 0 && divLow === 0) {
            divLow = 4;
        }

        this.config.clkPhase = clkPhase;
        this.config.clkPolarity = clkPolarity;
        this.config.bitDir = bitDir;
        this.config.charLen = charLen;
        this.config.csType = csType;
        this.config.csPolarity = csPolarity;
        this.config.divHigh = divHigh;
        this.config.divLow = divLow;
    };
    
    CustomSPI_Interface.prototype.configure = function(u2a) {
        this.enable(u2a);
        
        u2a.onPayloadPacketReceived = this.onPayloadPacketReceived;
        u2a.shouldPauseTransmission = this.shouldPauseTransmission;

        for(var i=0; i<this.transferPromises.length; i++) {
            this.transferPromises[i].reject();
        }
        this.cmdQueue.splice(0, this.cmdQueue.length);
        this.readyToProcess = true;
    };

    CustomSPI_Interface.prototype.enable = function(u2a) {
        console.log('SPI', 'Config', 'Params', [this.config.clkPhase, this.config.clkPolarity, this.config.bitDir, this.config.charLen, this.config.csType, this.config.csPolarity, this.config.divHigh, this.config.divLow]);
        u2a.sendCommandPacket(Cmd_SPI_Control,
            [this.config.clkPhase, this.config.clkPolarity, this.config.bitDir, this.config.charLen, this.config.csType, this.config.csPolarity, this.config.divHigh, this.config.divLow]
        );
    };

    CustomSPI_Interface.prototype.disable = function(u2a) {
    };

    CustomSPI_Interface.prototype.calculateWriteData = function(writeData) {
        var bytesToRead = 0, bytesToReadHigh, bytesToReadLow, waitTime = 0, waitTimeHigh, waitTimeLow, devAddr;
        var initByte = writeData[0];
        var noOfDevices = this.noOfDevices;

        if(initByte === 0x80) {
            devAddr = writeData[1];
            bytesToRead = writeData[4] + 7;
            // if(writeData[1] === 0x00) {
            //     waitTime = 100 + (bytesToRead * 10);
            // } else {
            //     waitTime = ((devAddr - 1) * 6) + (bytesToRead * 10) + 100;
            // }
        } else if(initByte === 0xA0) {
            bytesToRead = (writeData[3] + 7) * (noOfDevices > 0 ? noOfDevices - 1 : 0);
            // waitTime = ((noOfDevices > 2 ? noOfDevices - 2 : 0) * 6) + (bytesToRead * 10) + 100;
        } else if(initByte === 0xC0) {
            bytesToRead = (writeData[3] + 7) * noOfDevices;
            // waitTime = ((noOfDevices - 1) * 6) + (bytesToRead * 10) + 100;
        }

        bytesToReadHigh = Math.floor(bytesToRead / 256);
        bytesToReadLow = bytesToRead % 256;

        waitTime = 10000;
        waitTimeHigh = Math.floor(waitTime / 256);
        waitTimeLow = waitTime % 256;

        writeData.unshift(writeData.length);
        writeData.push(bytesToReadHigh);
        writeData.push(bytesToReadLow);
        writeData.push(waitTimeHigh);
        writeData.push(waitTimeLow);

        return writeData;
    };

    CustomSPI_Interface.prototype.shouldPauseTransmission = function(txPacket) {
        return false;
    };

    CustomSPI_Interface.prototype.onPayloadPacketReceived = function(target, cmd, rawDataSliced, rawData, payloadLen, payloadStart) {
        // console.log('SPI', 'onPayloadPacketReceived', target, cmd, rawData);
        if(cmd === 144 || cmd === 146) {
            var transferPromises = this.getInterfaces()['customspi'].transferPromises;
            // console.log('SPI', 'Transfer data in buffer', transferPromises.length);
            if (transferPromises && transferPromises.length > 0) {
                // console.log('SPI', 'Transfer promise resolve');
                var transferPromise = transferPromises.shift();
                // console.log('SPI', 'Transfer promise shifted', transferPromises.length);
                transferPromise.resolve({rawData: rawData, payloadLen: payloadLen, payloadStart: payloadStart});
            }
        } else if(cmd === 148) {
            // console.log('SPI', 'onPayloadPacketReceived', target, cmd, rawData, payloadLen);
            var status = rawData.slice(payloadStart, payloadStart + rawData[payloadLen])[0];
            if(status > 0) {
                // console.log('SPI', 'onPayloadPacketReceived', 'Timeout');
                // this.sendCommandPacket(148, this.getInterfaces()['customspi'].calculateWriteData([0x00]));
                this.sendCommandPacket(9, [1, 0x00]);
            }
            var spiInterface = this.getInterfaces()['customspi'];
            if(spiInterface.opPromise) {
                // console.log('SPI', 'onPayloadPacketReceived', 'Resolve');
                spiInterface.opPromise.resolve();
                spiInterface.opPromise = undefined;
            }
        } else {
            gc.console.debug('USB2ANY', 'Ignoring Payload Packet for command #', cmd);
        }
    };

    CustomSPI_Interface.prototype.recursivePing = function(u2a, count) {
        var deferred = Q.defer();

        // Clear Rx Buffer
        u2a.sendCommandPacket(147, []);

        // 606
        // var writeData = [0x80, count, 0x01, 0x04, 0x00];

        // 616
        var writeData = [0x80, count, 0x03, 0x06, 0x00];

        var crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
        writeData.push.apply(writeData, crc);
        writeData = this.calculateWriteData(writeData);

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        opPromise.promise.then(function() {
            var buffLenDeferred = Q.defer();
            this.transferPromises.push(buffLenDeferred);
            buffLenDeferred.promise.then(function(data) {
                var rawData = data.rawData;
                var payloadLen = data.payloadLen;
                var payloadStart = data.payloadStart;
                var payload = rawData.slice(payloadStart, payloadStart + rawData[payloadLen]);
                var dataLen = payload[1] * 256 + payload[0];

                if(dataLen > 0) {
                    this.recursivePing(u2a, count + 1).then(function(countData) {
                        deferred.resolve(countData);
                    }.bind(this));
                } else {
                    deferred.resolve(count);
                }
            }.bind(this));
            u2a.sendCommandPacket(146, []);
        }.bind(this));

        u2a.sendCommandPacket(148, writeData);

        return deferred.promise;
    }

    CustomSPI_Interface.prototype.recursiveRead = function(u2a, cmdArr, deferred, readData) {
        var cmd = cmdArr.shift();
        var noOfRegs = (cmd[0] < 160 ? cmd[4] : cmd[3]) + 1;
        var noOfDevices = cmd[0] < 160 ? 1 : this.noOfDevices;

        cmd = this.calculateWriteData(cmd);
        // console.log('SPI', 'Recursive read', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        opPromise.promise.then(function() {
            if(cmdArr.length > 0) {
                this.recursiveRead(u2a, cmdArr, deferred, readData);
            } else {
                var bufferLenDeferred = Q.defer();
                this.transferPromises.push(bufferLenDeferred);
                bufferLenDeferred.promise.then(function(data) {
                    // console.log('SPI', 'Recursive read', 'Buffer data length', data);
                    var rawData = data.rawData;
                    var payloadLen = data.payloadLen;
                    var payloadStart = data.payloadStart;
                    var payload = rawData.slice(payloadStart, payloadStart + rawData[payloadLen]);

                    var dataLen = payload[1] * 256 + payload[0];
                    var packetSize = Math.ceil(dataLen / 54);

                    function packetsTransferResolve(packet) {
                        // console.log('SPI', 'Recursive read', 'Packet', packet);
                        var packetPayloadLen = packet.rawData[packet.payloadLen];
                        var packetPayload = packet.rawData.slice(packet.payloadStart, packet.payloadStart + packetPayloadLen);
                        readData.push.apply(readData, packetPayload);
                        // console.log('SPI', 'Recursive read', 'Res data', readData);
                    }

                    var packetDeferred;
                    var packetsTransferPromise = [];
                    for(var i=0; i<packetSize; i++) {
                        packetDeferred = Q.defer();
                        packetsTransferPromise[i] = packetDeferred.promise;
                        this.transferPromises.push(packetDeferred);
                        packetsTransferPromise[i].then(packetsTransferResolve);
                    }
                    // console.log('SPI', 'Recursive read', 'Packet transfer promises', packetsTransferPromise);
                    Promise.all(packetsTransferPromise).then(function() {
                        // console.log('SPI', 'Recursive read', 'Final res', readData);

                        deferred.resolve(readData);
                    }.bind(this));

                    if(packetSize > 0) {
                        // console.log('SPI', 'Recursive read', 'Buffer data transfer', (new Date()).getTime());
                        u2a.sendCommandPacket(144, [payload[1], payload[0]]);
                    }

                }.bind(this));
                // console.log('SPI', 'Recursive read', 'Get buffer size', 'Timestamp', (new Date()).getTime());
                u2a.sendCommandPacket(146, []);
            }
        }.bind(this));

        u2a.sendCommandPacket(148, cmd);
    };

    CustomSPI_Interface.prototype.manageReads = function(u2a, cmdArr) {
        var deferred = Q.defer();
        var readData = [];

        // Clear Rx Buffer
        u2a.sendCommandPacket(147, []);

        if(cmdArr.length > 0) {
            this.recursiveRead(u2a, cmdArr, deferred, readData);
        } else {
            deferred.resolve(readData);
        }
        return deferred.promise;
    };

    CustomSPI_Interface.prototype.recursiveWrite = function(u2a, cmdArr, deferred) {
        var cmd = cmdArr.shift();
        // console.log('SPI', 'Recursive write', 'Timestamp', (new Date()).getTime());
        cmd = this.calculateWriteData(cmd);
        console.log('SPI', 'Recursive write', cmd);
        // console.log('SPI', 'Custom interface', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        opPromise.promise.then(function() {
            if(cmdArr.length > 0) {
                this.recursiveWrite(u2a, cmdArr, deferred);
            } else {
                deferred.resolve();
            }
        }.bind(this));

        u2a.sendCommandPacket(148, cmd);
    };

    CustomSPI_Interface.prototype.manageWrites = function(u2a, cmdArr) {
        var deferred = Q.defer();
        if(cmdArr.length > 0) {
            this.recursiveWrite(u2a, cmdArr, deferred);
        } else {
            deferred.resolve();
        }
        return deferred.promise;
    };

    CustomSPI_Interface.prototype.manageReadWrite = function(u2a, cmdInfo) {
        var regPage = document.querySelector('ti-widget-register-page');

        if(cmdInfo) {
            this.cmdQueue.push(cmdInfo);
        }
        // console.log('SPI', 'Manage', this.cmdQueue, u2a);
        if(this.readyToProcess && this.cmdQueue.length > 0) {
            var cmd = this.cmdQueue.shift();
            // console.log('SPI', 'Manage', 'Cmd', cmd);
            switch(cmd.type) {

                case 'write':
                    console.log('SPI', 'Manage', 'Write', cmd);

                    this.readyToProcess = false;
                    this.manageWrites(u2a, [
                        cmd.data
                    ]).then(function() {
                        console.log('SPI', 'Manage', 'Write', 'Resolved');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'multipleWrite':
                    console.log('SPI', 'Manage', 'Multiple write', cmd);

                    this.readyToProcess = false;
                    this.manageWrites(u2a, cmd.cmdArr).then(function() {
                        if(cmd.promise) {
                            cmd.promise.resolve();
                        }

                        console.log('SPI', 'Manage', 'Multiple write', 'Resolved');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'commClear':
                    console.log('SPI', 'Manage', 'Comm clear', u2a);

                    this.readyToProcess = false;
                    u2a.sendCommandPacket(9, [1, 0x00]);
                    this.readyToProcess = true;
                    this.manageReadWrite(u2a);
                    break;

                case 'read':
                    console.log('SPI', 'Manage', 'Read', cmd);
                    this.readyToProcess = false;

                    if(typeof cmd.data[0] === 'number') {
                        cmd.data = [cmd.data];
                    }

                    this.manageReads(u2a, cmd.data).then(function(resData) {
                        cmd.promise.resolve(resData);

                        console.log('SPI', 'Manage', 'Read', 'Resolve');

                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));
                    break;

                case 'poll':
                    console.log('SPI', 'Manage', 'Poll', cmd);
                    this.readyToProcess = false;

                    this.manageReads(u2a, [
                        [0xC0, 0x05, 0x36, 0x00, 0x3B, 0x85],  // System Faults
                        [0xC0, 0x05, 0x56, 0x01, 0xD2, 0x45],  // Cell - Balancing Status
                        [0xC0, 0x05, 0xB6, 0x01, 0x9B, 0x85],  // BAT Voltage
                        [0xC0, 0x05, 0x2B, 0x02, 0xB3, 0x14],  // Balancing, Device Status & Fault Summary
                        [0xC0, 0x05, 0x30, 0x02, 0xB9, 0xE4],  // Communication Faults 1, 2, 3
                        [0xC0, 0x05, 0x3C, 0x03, 0x7D, 0x24],  // Cells OVUV
                        [0xC0, 0x05, 0xAE, 0x03, 0x10, 0x44],  // Die Temperature
                        [0xC0, 0x03, 0x18, 0x0F, 0x86, 0x20],  // Cell - Balancing Timer
                        [0xC0, 0x05, 0x8E, 0x0F, 0x09, 0x81]  // GPIO Voltage
                    ]).then(function(resData1) {

                        this.manageReads(u2a, [
                            [0xC0, 0x05, 0x68, 0x21, 0xC3, 0xFD],  // Cell Voltage
                            [0xC0, 0x05, 0xB2, 0x23, 0x19, 0x5C],  // Diagnostics
                            [0xC0, 0x05, 0x8C, 0x01, 0x89, 0x25],  // TSREF
                            [0xC0, 0x05, 0xD6, 0x02, 0xF3, 0x84]   //CS 24bit
                        ]).then(function(resData2) {
                            resData1.push.apply(resData1, resData2);
                            cmd.promise.resolve(resData1);

                            console.log('SPI', 'Manage', 'Poll', 'Resolve');
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'poll-600':
                    console.log('SPI', 'Manage', 'Poll - 600', cmd);
                    this.readyToProcess = false;

                    this.manageReads(u2a, [
                        [0xA0, 0x05, 0x36, 0x00, 0x25, 0x85],  // System Faults
                        [0xA0, 0x05, 0x56, 0x01, 0xCC, 0x45],  // Cell - Balancing Status
                        [0xA0, 0x05, 0xB6, 0x01, 0x85, 0x85],  // BAT Voltage
                        [0xA0, 0x05, 0x2B, 0x02, 0xAD, 0x14],  // Balancing, Device Status & Fault Summary
                        [0xA0, 0x05, 0x30, 0x02, 0xA7, 0xE4],  // Communication Faults 1, 2, 3
                        [0xA0, 0x05, 0x3C, 0x03, 0x63, 0x24],  // Cells OVUV
                        [0xA0, 0x05, 0xAE, 0x03, 0x0E, 0x44],  // Die Temperature
                        [0xA0, 0x03, 0x18, 0x0F, 0x98, 0x20],  // Cell - Balancing Timer
                        [0xA0, 0x05, 0x8E, 0x0F, 0x17, 0x81]  // GPIO Voltage
                    ]).then(function(resData1) {

                        this.manageReads(u2a, [
                            [0xA0, 0x05, 0x68, 0x21, 0xDD, 0xFD],  // Cell Voltage
                            [0xA0, 0x05, 0xB2, 0x23, 0x07, 0x5C],  // Diagnostics
                            [0xA0, 0x05, 0x8C, 0x01, 0x97, 0x25],  // TSREF
                            [0xA0, 0x07, 0x80, 0x00, 0xF2, 0x25],  // Debug Comm Stat
                            [0x80, 0x00, 0x21, 0x00, 0x00, 0x75, 0xD4],  // 600 - Fault Summary
                            [0x80, 0x00, 0x23, 0x00, 0x00, 0xD4, 0x14],  // 600 - DEBUG_COMM_STAT
                            [0xA0, 0x05, 0xD6, 0x02, 0xED, 0x84]   //CS 24bit
                        ]).then(function(resData2) {
                            resData1.push.apply(resData1, resData2);
                            cmd.promise.resolve(resData1);

                            console.log('SPI', 'Manage', 'Poll - 600', 'Resolve');
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                        }.bind(this));

                    }.bind(this));
                    break;

                case 'autoAddress':
                    console.log('SPI', 'Manage', 'Auto address', cmd);
                    this.readyToProcess = false;

                    // 616
                    var writeData, crc;
                    // Broadcast 0x00 for ECC_TEST 0x34C
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    u2a.sendCommandPacket(148, writeData);
                    // Broadcast 0x01 for CONTROL1 0x309
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                    u2a.sendCommandPacket(148, writeData);
                    // Broadcast addr for DIR0_ADDR 0x306
                    for(var i=0; i<36; i++) {
                        writeData = [0xD0, 0x03, 0x06, i];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        writeData = this.calculateWriteData(writeData);
                        u2a.sendCommandPacket(148, writeData);
                    }
                    this.recursivePing(u2a, 0).then(function(noOfDevices) {
                        console.log('SPI', 'No. of devices', noOfDevices);
                        this.noOfDevices = noOfDevices;

                        // Broadcast 0x02 for COMM_CTRL 0x308
                        writeData = this.calculateWriteData([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                        u2a.sendCommandPacket(148, writeData);
                        if(noOfDevices > 1) {
                            // Single device 0x00 for COMM_CTRL 0x308 at 0x00
                            writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                            u2a.sendCommandPacket(148, writeData);
                            // Single device 0x03 for COMM_CTRL 0x308 at stack top
                            writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                            crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                            writeData.push.apply(writeData, crc);
                            writeData = this.calculateWriteData(writeData);
                            u2a.sendCommandPacket(148, writeData);
                        } else {
                            // Single device 0x01 for COMM_CTRL 0x308 at 0x00
                            writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                            u2a.sendCommandPacket(148, writeData);
                        }
                        this.manageReads(u2a, [
                            [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],  // Broadcast read ECC_TEST 0x34C
                            [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]  // Broadcast read DIR0_ADDR 0x306
                        ]).then(function() {
                            // Broadcast 0x03 for FAULT_RST2 0x332
                            writeData = this.calculateWriteData([0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                            u2a.sendCommandPacket(148, writeData);
                            // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                            writeData = this.calculateWriteData([0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);
                            u2a.sendCommandPacket(148, writeData);

                            cmd.promise.resolve(noOfDevices);

                            console.log('SPI', 'Manage', 'Auto address', 'Resolve');
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);
                        }.bind(this));

                    }.bind(this));
                    break;

                case 'reverse-autoaddress':
                    console.log('SPI', 'Manage', 'Reverse auto address', cmd);
                    this.readyToProcess = false;

                    var writeData, crc;
                    // Broadcast COMM_CTRL reg with value 0x00
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x08, 0x00, 0xCF, 0x24]);
                    u2a.sendCommandPacket(148, writeData);
                    // Write [DIR_SEL] = 1 to base device
                    writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    u2a.sendCommandPacket(148, writeData);
                    //Single device write to 600 CONTROL1[SEND_WAKE]=1 (wake up stack devices)
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x09, 0xA0, 0x12, 0x35]);
                    u2a.sendCommandPacket(148, writeData);

                    // setTimeout(function() {
                        // reverse broadcast [DIR_SEL] = 1
                        writeData = this.calculateWriteData([0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                        u2a.sendCommandPacket(148, writeData);
                        // broadcast [DIR_SEL] = 1 & [ADDR_WR] = 1
                        writeData = this.calculateWriteData([0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                        u2a.sendCommandPacket(148, writeData);

                        // broadcast DIR1_ADDR register = number of devices
                        for(var dev = 0; dev < 36; dev ++) {
                            writeData = [0xD0, 0x03, 0x07, dev];
                            crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                            writeData.push.apply(writeData, crc);
                            writeData = this.calculateWriteData(writeData);
                            u2a.sendCommandPacket(148, writeData);
                        }

                        this.recursivePing(u2a, 0).then(function(noOfDevices) {
                            console.log('Reverse Auto Address', 'No. of Devices', noOfDevices);
                            this.noOfDevices = noOfDevices;

                            // broadcast [STACK_DEV] = 1 - this will define the devices as a stack device
                            writeData = this.calculateWriteData([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            u2a.sendCommandPacket(148, writeData);
                            if(noOfDevices > 1) {
                                writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                                u2a.sendCommandPacket(148, writeData);

                                writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                writeData = this.calculateWriteData(writeData);
                                u2a.sendCommandPacket(148, writeData);
                            }
                            else {
                                writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                                u2a.sendCommandPacket(148, writeData);
                            }
                            this.manageReads(u2a, [
                                [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],   // Broadcast read ECC_TEST 0x34C
                                [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]    // Broadcast read DIR0_ADDR 0x306
                            ]).then(function() {
                                // Broadcast 0x03 for FAULT_RST2 0x332
                                writeData = this.calculateWriteData([0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                                u2a.sendCommandPacket(148, writeData);
                                // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                                writeData = this.calculateWriteData([0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);
                                u2a.sendCommandPacket(148, writeData);

                                cmd.promise.resolve(noOfDevices);

                                console.log('SPI', 'Manage', 'Reverse auto address', 'Resolve');
                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }.bind(this));

                        }.bind(this));

                    // }.bind(this), 100);
                    break;

                case 'autoAddress-600':
                    console.log('SPI', 'Manage', 'Auto address - 600', cmd);
                    this.readyToProcess = false;

                    // 600
                    var writeData, writeDataArr = [], crc;
                    // Single device 0x14 for DEV_CONF1 0x2001 at 0x00
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    // u2a.sendCommandPacket(148, writeData);
                    writeDataArr.push([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    // Single device 0x20 for CONTROL1 0x0309 at 0x00
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x09, 0x20, 0x13, 0x95]);
                    // u2a.sendCommandPacket(148, writeData);
                    // writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0x20, 0x13, 0x95]);

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        // setTimeout(function() {
                            writeDataArr = [];
                            // Broadcast 0x00 for ECC_TEST 0x34C
                            // writeData = this.calculateWriteData([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            // Broadcast 0x01 for CONTROL1 0x0309
                            // writeData = this.calculateWriteData([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                            // Broadcast addr for DIR0_ADDR 0x306
                            for(var i=0; i<36; i++) {
                                writeData = [0xD0, 0x03, 0x06, i];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // writeData = this.calculateWriteData(writeData);
                                // u2a.sendCommandPacket(148, writeData);
                                writeDataArr.push(writeData);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.recursivePing(u2a, 0).then(function(noOfDevices) {
                                    console.log('SPI', 'No. of devices', noOfDevices);
                                    this.noOfDevices = noOfDevices;

                                    writeDataArr = [];
                                    // Single device write to devices 1, 2, 3… address 0x308 = 0x02
                                    for(var dev = 1; dev < noOfDevices - 1; dev ++) {
                                        writeData = [0x90, dev, 0x03, 0x08, 0x02];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // writeData = this.calculateWriteData(writeData);
                                        // u2a.sendCommandPacket(148, writeData);
                                        writeDataArr.push(writeData);
                                    }
                                    if(noOfDevices > 1) {
                                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // writeData = this.calculateWriteData(writeData);
                                        // u2a.sendCommandPacket(148, writeData);
                                        writeDataArr.push(writeData);
                                    }

                                    this.manageWrites(u2a, writeDataArr).then(function() {

                                        this.manageReads(u2a, [
                                            [0xA0, 0x03, 0x4C, 0x00, 0xE6, 0xE4],  // Stack read ECC_TEST 0x306
                                            [0x80, 0x00, 0x20, 0x01, 0x00, 0x25, 0x84],  // Single read read DEV_CONF1 0x2001 at 0x00
                                            [0xA0, 0x03, 0x06, 0x00, 0xD1, 0x84]  // Stack read DIR0_ADDR 0x306
                                        ]).then(function() {
                                            writeDataArr = [];
                                            // Stack 0x03 for FAULT_RST2 0x332
                                            // writeData = this.calculateWriteData([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // Stack 0x00 for COMM_TIMEOUT_CONF 0x019
                                            // writeData = this.calculateWriteData([0xB0, 0x00, 0x13, 0x00, 0x2B, 0xD4]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0xB0, 0x00, 0x19, 0x00, 0x2D, 0x74]);
                                            // Single device 0x00 for COMM_TIMEOUT 0x2005 at 0x00
                                            // writeData = this.calculateWriteData([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);

                                            this.manageWrites(u2a, writeDataArr).then(function() {

                                                cmd.promise.resolve(noOfDevices);

                                                console.log('SPI', 'Manage', 'Auto address- 600', 'Resolve');
                                                this.readyToProcess = true;
                                                this.manageReadWrite(u2a);

                                            }.bind(this));

                                        }.bind(this));

                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        // }.bind(this), 100); // 35 (max no. of 616 stack) * 2.72 (time from shutdown to active + send wake-up)

                    }.bind(this));
                    break;

                case "reverse-autoaddress-600":
                    console.log('SPI', 'Manage', 'Reverse auto address - 600', cmd);
                    this.readyToProcess = false;

                    var writeData, writeDataArr = [], crc;
                    //Single device write to 600 CONTROL1[DIR_SEL]=1 
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    // u2a.sendCommandPacket(148, writeData);
                    writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0x80, 0x13, 0xED]);
                    //Single device write to 600 DEV_CONF1[SEL_TX_TONE]=0    
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    // u2a.sendCommandPacket(148, writeData);
                    writeDataArr.push([0x90, 0x00, 0x20, 0x01, 0x14, 0xE4, 0x48]);
                    //Single device write to 600 CONTROL1[SEND_WAKE]=1 (wake up stack devices)
                    // writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x09, 0xA0, 0x12, 0x35]);
                    // u2a.sendCommandPacket(148, writeData);
                    // writeDataArr.push([0x90, 0x00, 0x03, 0x09, 0xA0, 0x12, 0x35]);

                    this.manageWrites(u2a, writeDataArr).then(function() {

                        // setTimeout(function() {
                            //Broadcast write 0x00 for OTC_ECC_TEST 0x34C(dummy write, intend to sync up internal DLL)
                            // writeData = this.calculateWriteData([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                            //Broadcast write reverse address 0x309 = 0x80 (change stack devices direction DIR_SEL=1)
                            // writeData = this.calculateWriteData([0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xE0, 0x03, 0x09, 0x80, 0xC0, 0x14]);
                            //Broadcast write address 0x308 = 0x02
                            // writeData = this.calculateWriteData([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                            //Broadcast write address 0x309 = 0x81 (enable 616 auto addressing)
                            // writeData = this.calculateWriteData([0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                            // u2a.sendCommandPacket(148, writeData);
                            writeDataArr.push([0xD0, 0x03, 0x09, 0x81, 0x0E, 0xD4]);
                            //Broadcast write consecutively address 0x307 = 0, 1, 2, 3… (0 base, rest is stack)
                            for(var i=0; i<66; i++) {
                                writeData = [0xD0, 0x03, 0x07, i];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                writeData.push.apply(writeData, crc);
                                // writeData = this.calculateWriteData(writeData);
                                // u2a.sendCommandPacket(148, writeData);
                                writeDataArr.push(writeData);
                            }

                            this.manageWrites(u2a, writeDataArr).then(function() {

                                this.recursivePing(u2a, 0).then(function(noOfDevices) {
                                    console.log('SPI', 'No. of devices', noOfDevices);
                                    this.noOfDevices = noOfDevices;

                                    // Single device write to devices 1, 2, 3… address 0x308 = 0x02
                                    for(var dev = 1; dev < noOfDevices; dev ++) {
                                        writeData = [0x90, dev, 0x03, 0x08, 0x02];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // writeData = this.calculateWriteData(writeData);
                                        // u2a.sendCommandPacket(148, writeData);
                                        writeDataArr.push(writeData);
                                    }
                                    if(noOfDevices > 1) {
                                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                                        writeData.push.apply(writeData, crc);
                                        // writeData = this.calculateWriteData(writeData);
                                        // u2a.sendCommandPacket(148, writeData);
                                        writeDataArr.push(writeData);
                                    }

                                    this.manageWrites(u2a, writeDataArr).then(function() {

                                        this.manageReads(u2a, [
                                            [0xA0, 0x03, 0x4C, 0x00, 0xE6, 0xE4],  // Stack read OTC_ECC_TEST  0x34C
                                            [0xA0, 0x03, 0x07, 0x00, 0xD0, 0x14],  // Stack read address 0x307 (read back to verify addresses are correct)
                                            [0x80, 0x00, 0x20, 0x01, 0x00, 0x25, 0x84]  // Single read DEV_CONF1 0x2001 at 0x00
                                        ]).then(function() {
                                            // Stack 0x03 for FAULT_RST2 0x332
                                            // writeData = this.calculateWriteData([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0xB0, 0x03, 0x32, 0x03, 0x83, 0x85]);
                                            // Stack 0x00 for COMM_TIMEOUT_CONF 0x019
                                            // writeData = this.calculateWriteData([0xB0, 0x00, 0x13, 0x00, 0x2B, 0xD4]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0xB0, 0x00, 0x19, 0x00, 0x2D, 0x74]);
                                            // Single device 0x00 for COMM_TIMEOUT 0x2005 at 0x00
                                            // writeData = this.calculateWriteData([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);
                                            // u2a.sendCommandPacket(148, writeData);
                                            writeDataArr.push([0x90, 0x00, 0x20, 0x05, 0x00, 0xE6, 0x87]);

                                            this.manageWrites(u2a, writeDataArr).then(function() {

                                                cmd.promise.resolve(noOfDevices);

                                                console.log('SPI', 'Manage', 'Reverse auto address - 600', 'Resolve');
                                                this.readyToProcess = true;
                                                this.manageReadWrite(u2a);

                                            }.bind(this));

                                        }.bind(this));

                                    }.bind(this));

                                }.bind(this));

                            }.bind(this));

                        // }.bind(this), 100); // 35 (max no. of 616 stack) * 2.72 (time from shutdown to active + send wake-up)

                    }.bind(this));
                    break;

                case 'debugAutoAddress':
                    console.log('SPI', 'Manage', 'Debug auto address', cmd);
                    this.readyToProcess = false;

                    // 616
                    var writeData, crc;
                    var noOfDevices = cmd.noOfDevices;
                    // Broadcast 0x00 for ECC_TEST 0x34C
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x4C, 0x00, 0xFC, 0x24]);
                    u2a.sendCommandPacket(148, writeData);
                    // Broadcast 0x01 for CONTROL1 0x309
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x09, 0x01, 0x0F, 0x74]);
                    u2a.sendCommandPacket(148, writeData);
                    // Broadcast addr for DIR0_ADDR 0x306
                    for(var i=0; i<noOfDevices; i++) {
                        writeData = [0xD0, 0x03, 0x06, i];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        writeData = this.calculateWriteData(writeData);
                        u2a.sendCommandPacket(148, writeData);
                    }
                    // Broadcast 0x02 for COMM_CTRL 0x308
                    writeData = this.calculateWriteData([0xD0, 0x03, 0x08, 0x02, 0x4E, 0xE5]);
                    u2a.sendCommandPacket(148, writeData);
                    if(noOfDevices > 1) {
                        // Single device 0x00 for COMM_CTRL 0x308 at 0x00
                        writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x00, 0x13, 0xDD]);
                        u2a.sendCommandPacket(148, writeData);
                        // Single device 0x03 for COMM_CTRL 0x308 at stack top
                        writeData = [0x90, noOfDevices - 1, 0x03, 0x08, 0x03];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                        writeData.push.apply(writeData, crc);
                        writeData = this.calculateWriteData(writeData);
                        u2a.sendCommandPacket(148, writeData);
                    } else {
                        // Single device 0x01 for COMM_CTRL 0x308 at 0x00
                        writeData = this.calculateWriteData([0x90, 0x00, 0x03, 0x08, 0x01, 0xD2, 0x1D]);
                        u2a.sendCommandPacket(148, writeData);
                    }
                    this.manageReads(u2a, [
                        [0xC0, 0x03, 0x4C, 0x00, 0xF8, 0xE4],  // Broadcast read ECC_TEST 0x34C
                        [0xC0, 0x03, 0x06, 0x00, 0xCF, 0x84]  // Broadcast read DIR0_ADDR 0x306
                    ]).then(function(resData) {
                        // console.log('SPI', 'Debug auto address', 'Final data', resData);

                        // Broadcast 0x03 for FAULT_RST2 0x332
                        writeData = this.calculateWriteData([0xD0, 0x03, 0x32, 0x03, 0x9D, 0x85]);
                        u2a.sendCommandPacket(148, writeData);
                        // Broadcast 0x00 for COMM_TIMEOUT_CONF 0x019
                        writeData = this.calculateWriteData([0xD0, 0x00, 0x19, 0x00, 0x33, 0x74]);
                        u2a.sendCommandPacket(148, writeData);

                        cmd.promise.resolve({resData: resData, noOfDevices: cmd.noOfDevices});

                        console.log('SPI', 'Manage', 'Debug auto address', 'Resolve');
                        this.readyToProcess = true;
                        this.manageReadWrite(u2a);
                    }.bind(this));

                    break;

                case 'readAllBroadcast':
                    console.log('SPI', 'Manage', 'Read all broadcast', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128BroadcastRead() {
                        readData = [];

                        readData.push(0xC0);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('SPI', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('SPI', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128BroadcastRead.bind(this)();
                            } else {
                                console.log('SPI', 'Interface', 'Read all', 'Consoldated data', consolidatedData);

                                var regDataArr, regDataRowIndex, regDataColumnIndex, regAddr, rejectFlag;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128) * this.noOfDevices;
                                    regDataColumnIndex = regAddr % 128;
                                    regDataArr = [];
                                    rejectFlag = false;
                                    for(var j=0; j<this.noOfDevices; j++) {
                                        // console.log(regAddr, regDataRowIndex + j, regDataColumnIndex + 4);
                                        if(consolidatedData[regDataRowIndex + j]) {
                                            regDataArr.push(consolidatedData[regDataRowIndex + j][regDataColumnIndex + 4]);
                                        } else {
                                            rejectFlag = true;
                                        }
                                    }
                                    if(rejectFlag) {
                                        cmd.promiseArr[i].promise.reject();
                                    } else {
                                        cmd.promiseArr[i].promise.resolve(regDataArr.reverse());
                                        resolveCount++;
                                        console.log('SPI', 'Interface', 'Read all - Broadcast', cmd.promiseArr[i].addr, regDataArr);
                                    }
                                }

                                console.log('SPI', 'Interface', 'Read all - Broadcast', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('SPI', 'Manage', 'Read all broadcast', 'Regmap - Hide progress', regPage);

                                console.log('SPI', 'Manage', 'Read all broadcast', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128BroadcastRead.bind(this)();

                    break;

                case 'readAllStack':
                    console.log('SPI', 'Manage', 'Read all stack', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128StackRead() {
                        readData = [];

                        readData.push(0xA0);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('SPI', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('SPI', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128StackRead.bind(this)();
                            } else {
                                console.log('SPI', 'Interface', 'Read all - Stack', 'Consolidated data', consolidatedData);

                                var regDataArr, regDataRowIndex, regDataColumnIndex, regAddr, rejectFlag;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128) * (this.noOfDevices - 1);
                                    regDataColumnIndex = regAddr % 128;
                                    regDataArr = [];
                                    rejectFlag = false;
                                    for(var j=0; j<this.noOfDevices - 1; j++) {
                                        // console.log(regAddr, regDataRowIndex + j, regDataColumnIndex + 4);
                                        if(consolidatedData[regDataRowIndex + j]) {
                                            regDataArr.push(consolidatedData[regDataRowIndex + j][regDataColumnIndex + 4]);
                                        } else {
                                            rejectFlag = true;
                                        }
                                    }
                                    if(rejectFlag) {
                                        cmd.promiseArr[i].promise.reject();
                                    } else {
                                        cmd.promiseArr[i].promise.resolve(regDataArr.reverse());
                                        resolveCount++;
                                        console.log('SPI', 'Interface', 'Read all - Stack', cmd.promiseArr[i].addr, regDataArr);
                                    }
                                }

                                console.log('SPI', 'Interface', 'Read all - Stack', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('SPI', 'Manage', 'Read all stack', 'Regmap - Hide progress', regPage);

                                console.log('SPI', 'Manage', 'Read all stack', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128StackRead.bind(this)();

                    break;

                case 'readAllSingle':
                    console.log('SPI', 'Manage', 'Read all single', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 0, readData, crc;

                    function recursive128SingleRead() {
                        readData = [];

                        readData.push(0x80);
                        readData.push(cmd.device);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('SPI', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('SPI', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 2048) {
                                recursive128SingleRead.bind(this)();
                            } else {
                                console.log('SPI', 'Interface', 'Read all', 'Consoldated data', consolidatedData);

                                var regDataRowIndex, regDataColumnIndex, regAddr;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr;
                                    regDataRowIndex = Math.floor(regAddr/128);
                                    regDataColumnIndex = regAddr % 128;
                                    // console.log(regAddr, regDataRowIndex, regDataColumnIndex + 4);
                                    if(consolidatedData[regDataRowIndex]) {
                                        cmd.promiseArr[i].promise.resolve(consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                        resolveCount++;
                                        console.log('SPI', 'Interface', 'Read all - Single', cmd.promiseArr[i].addr, consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                    } else {
                                        cmd.promiseArr[i].promise.reject();
                                    }
                                }

                                console.log('SPI', 'Interface', 'Read all - Single', 'Done', resolveCount);

                                regReadDetail = undefined;

                                regPage.isUpdating = false;
                                console.log('SPI', 'Manage', 'Read all single', 'Regmap - Hide progress', regPage);

                                console.log('SPI', 'Manage', 'Read all single', 'Resolve');

                                this.readyToProcess = true;
                                this.manageReadWrite(u2a);
                            }
                        }.bind(this));
                    }

                    recursive128SingleRead.bind(this)();

                    break;

                case 'readAllSingle600':
                    console.log('SPI', 'Manage', 'Read all single - 600', cmd);
                    this.readyToProcess = false;

                    var consolidatedData = [], startAddr = 8192, readData, crc;

                    function recursive128SingleRead600() {
                        readData = [];

                        readData.push(0x80);
                        readData.push(cmd.device);
                        readData.push.apply(readData, getByteArrFromHexStr(prependZero(startAddr.toString(16), 4)));
                        readData.push(0x7F);
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        startAddr += 128;

                        this.manageReads(u2a, [readData]).then(function(resData) {
                            // console.log('SPI', 'Manage', 'Final res', resData);

                            var data = parseDeviceResponse(resData);
                            consolidatedData.push.apply(consolidatedData, data);

                            // console.log('SPI', 'Interface', 'Read all', 'Parsed data', data);

                            if(startAddr < 9216) {
                                recursive128SingleRead600.bind(this)();
                            } else {
                                console.log('SPI', 'Interface', 'Read all - 600', 'Consolidated data', consolidatedData);

                                var regDataRowIndex, regDataColumnIndex, regAddr;

                                var resolveCount = 0;

                                for(var i=0; i<cmd.promiseArr.length; i++) {
                                    regAddr = cmd.promiseArr[i].addr - 8192;
                                    regDataRowIndex = Math.floor(regAddr/128);
                                    regDataColumnIndex = regAddr % 128;
                                    // console.log(regAddr, regDataRowIndex, regDataColumnIndex + 4);
                                    if(consolidatedData[regDataRowIndex]) {
                                        cmd.promiseArr[i].promise.resolve(consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                        resolveCount++;
                                        console.log('SPI', 'Interface', 'Read all - 600', cmd.promiseArr[i].addr, consolidatedData[regDataRowIndex][regDataColumnIndex + 4]);
                                    } else {
                                        cmd.promiseArr[i].promise.reject();
                                    }
                                }

                                readData = [0x80, cmd.device, 0x03, 0x06, 0x04];
                                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                                readData.push.apply(readData, crc);

                                this.manageReads(u2a, [readData]).then(function(resData2) {
                                    // var data = parseDeviceResponse(resData2)[0];
                                    var data = resData2;
                                    console.log('SPI', 'Interface', 'Read all - 600', 'Consolidated data 2', data);
                                    var addr, val;

                                    for(var i=0; i<cmd.promiseArr.length; i++) {
                                        addr = cmd.promiseArr[i].addr;

                                        if(addr > 0x305 && addr < 0x30B) {
                                            val = data[addr - 0x306 + 4];
                                            if(val) {
                                                cmd.promiseArr[i].promise.resolve(val);
                                            } else {
                                                cmd.promiseArr[i].promise.reject();
                                            }
                                            resolveCount++;
                                            console.log('SPI', 'Interface', 'Read all - 600', cmd.promiseArr[i].addr, val);
                                        }
                                    }

                                    console.log('SPI', 'Interface', 'Read all - Single - 600', 'Done', resolveCount);

                                    regReadDetail = undefined;

                                    regPage.isUpdating = false;
                                    console.log('SPI', 'Manage', 'Read all single - 600', 'Regmap - Hide progress', regPage);

                                    console.log('SPI', 'Manage', 'Read all single - 600', 'Resolve');

                                    this.readyToProcess = true;
                                    this.manageReadWrite(u2a);
                                }.bind(this));

                            }
                        }.bind(this));
                    }

                    recursive128SingleRead600.bind(this)();

                    break;

                case 'cb-remaining-time':
                    var remainingTimeArr = [];
                    this.readyToProcess = false;
                    function readRemainingTime(cellNo) {
                      writeDataArr = [];
                      writeData = [0x90, cmd.device, 0x03, 0x30, (cellNo*2)+1];
                      crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
                      writeData.push.apply(writeData, crc);
                      writeDataArr.push(writeData);
                      this.manageWrites(u2a,writeDataArr).then(function () {

                        readData = [0x80, cmd.device, 0x05, 0x58, 0x00];
                        crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                        readData.push.apply(readData, crc);

                        this.manageReads(u2a, [readData]).then(function (data) {
                          remainingTimeArr.push(data[4]);
                          if(cellNo < 15) {
                            readRemainingTime.bind(this)(cellNo+1);
                          } else {
                            cmd.promise.resolve(remainingTimeArr);
                            remainingTimeArr = [];  
                            this.readyToProcess = true;
                            this.manageReadWrite(u2a);

                          }
                        }.bind(this));
                      }.bind(this));
                    }

                    readRemainingTime.bind(this)(0);
                    
                    break;
            }
        }
    }

    /**
     * Construct and send a command packet for reading from a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @return - A promise that is resolved once the data is received
     */
    CustomSPI_Interface.prototype.read = function (u2a, info, regModel, coreIndex) {
        console.log('SPI', 'Read', u2a, info, regModel, coreIndex, getDeviceIndexSpi(coreIndex, regModel._id), regReadDetail);
        var regAddr = info.addr;
        var readData = [], crc;

        var deviceIndex = getDeviceIndexSpi(coreIndex, regModel._id);

        var regPage = document.querySelector('ti-widget-register-page');
 
        if(regReadDetail) {
        // if(regReadDetail && ((regReadDetail.regModelId === regModel._id) && ((coreIndex === undefined) || (regReadDetail.coreIndex === coreIndex)))) {

            if(deviceIndex < 0) {
                var deferred = Q.defer();

                regReadPromiseArr.push({
                    promise: deferred,
                    addr: info.addr
                });

                regReadCount++;

                if(regReadCount === regReadDetail.numRegs) {
                    this.manageReadWrite(u2a, {
                        type: this.is600 === true ? (regModel._id === 'bq79600' ? 'readAllBroadcast600' : 'readAllStack') : 'readAllBroadcast',
                        device: deviceIndex,
                        count: regReadCount,
                        promiseArr: regReadPromiseArr
                    });
                }

                return deferred.promise;
            } else {
                var deferred = Q.defer();

                regReadPromiseArr.push({
                    promise: deferred,
                    addr: info.addr
                });

                regReadCount++;

                if(regReadCount === regReadDetail.numRegs) {
                    this.manageReadWrite(u2a, {
                        type: regModel._id === 'bq79600' ? 'readAllSingle600' : 'readAllSingle',
                        device: deviceIndex,
                        count: regReadCount,
                        promiseArr: regReadPromiseArr
                    });
                }

                return deferred.promise;
            }

        } else {

            regPage.isUpdating = true;
            console.log('SPI', 'Read', 'Regmap - Show progress', regPage);

            if(deviceIndex < 0) {
                readData.push(0xC0);    // Broadcast read

                regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
                readData.push.apply(readData, regAddr);    // Register address
                readData.push(0x00);    // Read 1 byte of data
                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                readData.push.apply(readData, crc);

                var deferred = Q.defer();

                var readPromise = Q.defer();
                readPromise.promise.then(function(data) {
                    var readData = [];
                    data = parseDeviceResponse(data);
                    for(var i=0; i<data.length; i++) {
                        readData.push(data[i][4]);
                    }
                    readData = readData.reverse();
                    deferred.resolve(readData);
                    regPage.isUpdating = false;
                    console.log('SPI', 'Read', 'Broadcast read', 'Regmap - Hide progress', regPage);
                })
                this.manageReadWrite(u2a, {type: 'read', data: readData, promise: readPromise});

                return deferred.promise;
            } else {
                readData.push(0x80);    // Single device read intialization byte
                readData.push(deviceIndex);    // Device address
                regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
                readData.push.apply(readData, regAddr);    // Register address
                readData.push(0x00);    // Read 1 byte of data
                crc = getByteArrFromHexStr(prependZero(getCRC16IBM(readData).toString(16), 4));
                readData.push.apply(readData, crc);    // CRC bytes

                var deferred = Q.defer();

                var readPromise = Q.defer();
                readPromise.promise.then(function(data) {
                    deferred.resolve(data[4]);
                    regPage.isUpdating = false;
                    console.log('SPI', 'Read', 'Single read', 'Regmap - Hide progress', regPage);
                })
                this.manageReadWrite(u2a, {type: 'read', data: readData, promise: readPromise});

                return deferred.promise;
            }

        }
    };

    /**
     * Construct and send a command packet for writing to a register
     *
     * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
     * @param info - The register map info for the register
     * @param value - The value to write to the register
     * @return - A promise that resolves immediately
     */
    CustomSPI_Interface.prototype.write = function (u2a, info, value, regModel, coreIndex) {
        console.log('SPI', 'Write', u2a, info, value, coreIndex);
        var writeData = [];
        var regAddr, crc;

        var deviceIndex = getDeviceIndexSpi(coreIndex, regModel._id);

        if(deviceIndex > -1) {
            writeData.push(0x90);    // Single device write intialization byte
            writeData.push(deviceIndex);    // Device address
            regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
            writeData.push.apply(writeData, regAddr);    // Register address
            writeData.push(value);    // Data byte
            crc = getByteArrFromHexStr(prependZero(getCRC16IBM(writeData).toString(16), 4));
            writeData.push.apply(writeData, crc);    // CRC bytes

            console.log('SPI', 'Registermap', 'Write', writeData);
            this.manageReadWrite(u2a, {type: 'write', data: writeData});
        }

        return Q(true);
    };

    var readRegsHdlr = function(eventType, event) {
        console.log('SPI', 'Reg read event', eventType, event, event.detail);

        var detailObj = event.detail;
        if (detailObj.type === undefined) {
            detailObj.type = eventType;
        }

        var modelId = undefined;
        if (detailObj.regModelId) {
            modelId = detailObj.regModelId;
        }

        var deviceIndex = getDeviceIndexSpi(detailObj.coreIndex, modelId);

        console.log("bc: read_regs: numRegs="+detailObj.numRegs+", regModel id="+modelId+", coreIndex="+detailObj.coreIndex+", type="+detailObj.type);
        if(detailObj.type === 'init' || detailObj.type === 'button' || detailObj.type === 'refresh') {
            // setTimeout(function() {
                // var modelId = "undefined";

                var regPage = document.querySelector('ti-widget-register-page');
                regPage.isUpdating = true;
                console.log('SPI', 'Reg read event', 'Regmap - Show progress', regPage);

                if (regReadDetail && regReadDetail.modelId) {
                    modelId = regReadDetail.modelId;
                }
                console.log("regReadDetail.regModelId changed from "+modelId +" to "+detailObj.regModelId);
                regReadDetail = detailObj;
                regReadDetail.device = deviceIndex;
                regReadCount = 0;
                regReadPromiseArr = [];
            // },0);
        }
    }

    var refreshPage = function() {
        var regPage = document.querySelector('ti-widget-register-page');
        regPage.readAllRegistersClickHdlr();
        templateObj.timerHdlr = window.setTimeout(refreshPage, templateObj.refreshInMs);
    };

    var templateObj = document.querySelector('#template_obj');
    templateObj.timerHdlr = null;
    templateObj.refreshInMs = 100;

    templateObj.addEventListener('dom-change', function(event) {
        if(document.querySelector('ti-bq79616-global')) {
            document.addEventListener("refresh-regs", readRegsHdlr.bind(templateObj,"refresh"));
            document.addEventListener("init-regs", readRegsHdlr.bind(templateObj,"init"));
            document.addEventListener("read_regs", readRegsHdlr.bind(templateObj,"statusOrButton"));
            document.addEventListener("read_reg", function(event) {
                regReadDetail = undefined;
            });
            document.addEventListener("auto_read_config_changed", function(event) {
               if(templateObj.timerHdlr) {
                   window.clearTimeout(templateObj.timerHdlr);
                   templateObj.timerHdlr = null;
               }
               var autoReadObj = event.detail.autoRead;
               if (autoReadObj.value > -1) {
                   templateObj.refreshInMs = +autoReadObj.value;
                   templateObj.timerHdlr = window.setTimeout(refreshPage, templateObj.refreshInMs);
               }
            });
        }
    });

};


// Wait until USB2ANY global var is ready before defining interface
if (gc.databind.USB2ANY) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init);
}